/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.hop.pipeline.transforms.http;

import java.util.List;
import lombok.Getter;
import lombok.Setter;
import org.apache.hop.core.CheckResult;
import org.apache.hop.core.Const;
import org.apache.hop.core.ICheckResult;
import org.apache.hop.core.annotations.Transform;
import org.apache.hop.core.encryption.Encr;
import org.apache.hop.core.exception.HopTransformException;
import org.apache.hop.core.exception.HopXmlException;
import org.apache.hop.core.row.IRowMeta;
import org.apache.hop.core.row.IValueMeta;
import org.apache.hop.core.row.value.ValueMetaInteger;
import org.apache.hop.core.row.value.ValueMetaString;
import org.apache.hop.core.util.Utils;
import org.apache.hop.core.variables.IVariables;
import org.apache.hop.core.xml.XmlHandler;
import org.apache.hop.i18n.BaseMessages;
import org.apache.hop.metadata.api.IHopMetadataProvider;
import org.apache.hop.pipeline.PipelineMeta;
import org.apache.hop.pipeline.transform.BaseTransformMeta;
import org.apache.hop.pipeline.transform.TransformMeta;
import org.w3c.dom.Node;

@Transform(
    id = "Http",
    image = "http.svg",
    name = "i18n::HTTP.Name",
    description = "i18n::HTTP.Description",
    categoryDescription = "i18n:org.apache.hop.pipeline.transform:BaseTransform.Category.Utility",
    keywords = "i18n::HttpMeta.keyword",
    documentationUrl = "/pipeline/transforms/http.html")
@Getter
@Setter
public class HttpMeta extends BaseTransformMeta<Http, HttpData> {
  private static final Class<?> PKG = HttpMeta.class;

  // the timeout for waiting for data (milliseconds)
  public static final int DEFAULT_SOCKET_TIMEOUT = 10000;

  // the timeout until a connection is established (milliseconds)
  public static final int DEFAULT_CONNECTION_TIMEOUT = 10000;

  // the time to wait till a connection is closed (milliseconds)? -1 is no not close.
  public static final int DEFAULT_CLOSE_CONNECTIONS_TIME = -1;
  public static final String CONST_HEADER = "header";
  public static final String CONST_RESULT = "result";
  public static final String CONST_SPACES_LONG = "        ";
  public static final String CONST_SPACES = "      ";
  public static final String CONST_PARAMETER = "parameter";

  private String socketTimeout;
  private String connectionTimeout;
  private String closeIdleConnectionsTime;

  /** URL / service to be called */
  private String url;

  /** function arguments : fieldname */
  private String[] argumentField;

  /** IN / OUT / INOUT */
  private String[] argumentParameter;

  /** function result: new value name */
  private String fieldName;

  /** The encoding to use for retrieval of the data */
  private String encoding;

  private boolean urlInField;

  private boolean ignoreSsl;

  private String urlField;

  private String proxyHost;

  private String proxyPort;

  private String httpLogin;

  private String httpPassword;

  private String resultCodeFieldName;
  private String responseTimeFieldName;
  private String responseHeaderFieldName;

  private String[] headerParameter;
  private String[] headerField;

  public HttpMeta() {
    super(); // allocate BaseTransformMeta
  }

  @Override
  public void loadXml(Node transformNode, IHopMetadataProvider metadataProvider)
      throws HopXmlException {
    readData(transformNode);
  }

  public void allocate(int nrargs, int nrqueryparams) {
    argumentField = new String[nrargs];
    argumentParameter = new String[nrargs];
    headerField = new String[nrqueryparams];
    headerParameter = new String[nrqueryparams];
  }

  @Override
  public Object clone() {
    HttpMeta retval = (HttpMeta) super.clone();
    int nrargs = argumentField.length;
    int nrheaderparams = headerField.length;

    retval.allocate(nrargs, nrheaderparams);

    System.arraycopy(argumentField, 0, retval.argumentField, 0, nrargs);
    System.arraycopy(argumentParameter, 0, retval.argumentParameter, 0, nrargs);
    System.arraycopy(headerField, 0, retval.headerField, 0, nrheaderparams);
    System.arraycopy(headerParameter, 0, retval.headerParameter, 0, nrheaderparams);

    return retval;
  }

  @Override
  public void setDefault() {
    socketTimeout = String.valueOf(DEFAULT_SOCKET_TIMEOUT);
    connectionTimeout = String.valueOf(DEFAULT_CONNECTION_TIMEOUT);
    closeIdleConnectionsTime = String.valueOf(DEFAULT_CLOSE_CONNECTIONS_TIME);
    int i;
    int nrargs;
    int nrquery;
    nrargs = 0;
    nrquery = 0;

    allocate(nrargs, nrquery);

    for (i = 0; i < nrargs; i++) {
      argumentField[i] = "arg" + i;
      argumentParameter[i] = "arg";
    }

    for (i = 0; i < nrquery; i++) {
      headerField[i] = CONST_HEADER + i;
      headerParameter[i] = CONST_HEADER;
    }

    fieldName = CONST_RESULT;
    resultCodeFieldName = "";
    responseTimeFieldName = "";
    responseHeaderFieldName = "";
    encoding = "UTF-8";
  }

  @Override
  public void getFields(
      IRowMeta inputRowMeta,
      String name,
      IRowMeta[] info,
      TransformMeta nextTransform,
      IVariables variables,
      IHopMetadataProvider metadataProvider)
      throws HopTransformException {
    if (!Utils.isEmpty(fieldName)) {
      IValueMeta v = new ValueMetaString(fieldName);
      v.setOrigin(name);
      inputRowMeta.addValueMeta(v);
    }
    if (!Utils.isEmpty(resultCodeFieldName)) {
      IValueMeta v = new ValueMetaInteger(variables.resolve(resultCodeFieldName));
      v.setOrigin(name);
      inputRowMeta.addValueMeta(v);
    }
    if (!Utils.isEmpty(responseTimeFieldName)) {
      IValueMeta v = new ValueMetaInteger(variables.resolve(responseTimeFieldName));
      v.setOrigin(name);
      inputRowMeta.addValueMeta(v);
    }
    String headerFieldName = variables.resolve(responseHeaderFieldName);
    if (!Utils.isEmpty(headerFieldName)) {
      IValueMeta v = new ValueMetaString(headerFieldName);
      v.setOrigin(name);
      inputRowMeta.addValueMeta(v);
    }
  }

  @Override
  public String getXml() {
    StringBuilder retval = new StringBuilder(300);

    retval.append("    ").append(XmlHandler.addTagValue("url", url));
    retval.append("    " + XmlHandler.addTagValue("urlInField", urlInField));
    retval.append("    " + XmlHandler.addTagValue("ignoreSsl", ignoreSsl));
    retval.append("    " + XmlHandler.addTagValue("urlField", urlField));
    retval.append("    " + XmlHandler.addTagValue("encoding", encoding));
    retval.append("    " + XmlHandler.addTagValue("httpLogin", httpLogin));
    retval.append(
        "    "
            + XmlHandler.addTagValue(
                "httpPassword", Encr.encryptPasswordIfNotUsingVariables(httpPassword)));
    retval.append("    " + XmlHandler.addTagValue("proxyHost", proxyHost));
    retval.append("    " + XmlHandler.addTagValue("proxyPort", proxyPort));
    retval.append("    " + XmlHandler.addTagValue("socketTimeout", socketTimeout));
    retval.append("    " + XmlHandler.addTagValue("connectionTimeout", connectionTimeout));
    retval.append(
        "    " + XmlHandler.addTagValue("closeIdleConnectionsTime", closeIdleConnectionsTime));

    retval.append("    <lookup>").append(Const.CR);

    for (int i = 0; i < argumentField.length; i++) {
      retval.append("      <arg>").append(Const.CR);
      retval.append(CONST_SPACES_LONG).append(XmlHandler.addTagValue("name", argumentField[i]));
      retval
          .append(CONST_SPACES_LONG)
          .append(XmlHandler.addTagValue(CONST_PARAMETER, argumentParameter[i]));
      retval.append("      </arg>").append(Const.CR);
    }
    for (int i = 0; i < headerField.length; i++) {
      retval.append("      <header>" + Const.CR);
      retval.append(CONST_SPACES_LONG + XmlHandler.addTagValue("name", headerField[i]));
      retval.append(
          CONST_SPACES_LONG + XmlHandler.addTagValue(CONST_PARAMETER, headerParameter[i]));
      retval.append("      </header>" + Const.CR);
    }

    retval.append("    </lookup>").append(Const.CR);

    retval.append("    <result>").append(Const.CR);
    retval.append(CONST_SPACES).append(XmlHandler.addTagValue("name", fieldName));
    retval.append(CONST_SPACES).append(XmlHandler.addTagValue("code", resultCodeFieldName));
    retval
        .append(CONST_SPACES)
        .append(XmlHandler.addTagValue("response_time", responseTimeFieldName));
    retval
        .append(CONST_SPACES)
        .append(XmlHandler.addTagValue("response_header", responseHeaderFieldName));
    retval.append("    </result>").append(Const.CR);

    return retval.toString();
  }

  private void readData(Node transformNode) throws HopXmlException {
    try {
      int nrargs;

      url = XmlHandler.getTagValue(transformNode, "url");
      urlInField = "Y".equalsIgnoreCase(XmlHandler.getTagValue(transformNode, "urlInField"));
      ignoreSsl = "Y".equalsIgnoreCase(XmlHandler.getTagValue(transformNode, "ignoreSsl"));
      urlField = XmlHandler.getTagValue(transformNode, "urlField");
      encoding = XmlHandler.getTagValue(transformNode, "encoding");
      httpLogin = XmlHandler.getTagValue(transformNode, "httpLogin");
      httpPassword =
          Encr.decryptPasswordOptionallyEncrypted(
              XmlHandler.getTagValue(transformNode, "httpPassword"));
      proxyHost = XmlHandler.getTagValue(transformNode, "proxyHost");
      proxyPort = XmlHandler.getTagValue(transformNode, "proxyPort");

      socketTimeout = XmlHandler.getTagValue(transformNode, "socketTimeout");
      connectionTimeout = XmlHandler.getTagValue(transformNode, "connectionTimeout");
      closeIdleConnectionsTime = XmlHandler.getTagValue(transformNode, "closeIdleConnectionsTime");

      Node lookup = XmlHandler.getSubNode(transformNode, "lookup");
      nrargs = XmlHandler.countNodes(lookup, "arg");

      int nrheaders = XmlHandler.countNodes(lookup, CONST_HEADER);
      allocate(nrargs, nrheaders);

      for (int i = 0; i < nrargs; i++) {
        Node anode = XmlHandler.getSubNodeByNr(lookup, "arg", i);

        argumentField[i] = XmlHandler.getTagValue(anode, "name");
        argumentParameter[i] = XmlHandler.getTagValue(anode, CONST_PARAMETER);
      }

      for (int i = 0; i < nrheaders; i++) {
        Node anode = XmlHandler.getSubNodeByNr(lookup, CONST_HEADER, i);
        headerField[i] = XmlHandler.getTagValue(anode, "name");
        headerParameter[i] = XmlHandler.getTagValue(anode, CONST_PARAMETER);
      }

      fieldName = XmlHandler.getTagValue(transformNode, CONST_RESULT, "name");
      resultCodeFieldName = XmlHandler.getTagValue(transformNode, CONST_RESULT, "code");
      responseTimeFieldName = XmlHandler.getTagValue(transformNode, CONST_RESULT, "response_time");
      responseHeaderFieldName =
          XmlHandler.getTagValue(transformNode, CONST_RESULT, "response_header");
    } catch (Exception e) {
      throw new HopXmlException(
          BaseMessages.getString(PKG, "HTTPMeta.Exception.UnableToReadTransformMeta"), e);
    }
  }

  @Override
  public void check(
      List<ICheckResult> remarks,
      PipelineMeta pipelineMeta,
      TransformMeta transformMeta,
      IRowMeta prev,
      String[] input,
      String[] output,
      IRowMeta info,
      IVariables variables,
      IHopMetadataProvider metadataProvider) {
    CheckResult cr;

    // See if we have input streams leading to this transform!
    if (input.length > 0) {
      cr =
          new CheckResult(
              ICheckResult.TYPE_RESULT_OK,
              BaseMessages.getString(PKG, "HTTPMeta.CheckResult.ReceivingInfoFromOtherTransforms"),
              transformMeta);
      remarks.add(cr);
    } else {
      cr =
          new CheckResult(
              ICheckResult.TYPE_RESULT_ERROR,
              BaseMessages.getString(PKG, "HTTPMeta.CheckResult.NoInpuReceived"),
              transformMeta);
      remarks.add(cr);
    }
    // check Url
    if (urlInField) {
      if (Utils.isEmpty(urlField)) {
        cr =
            new CheckResult(
                ICheckResult.TYPE_RESULT_ERROR,
                BaseMessages.getString(PKG, "HTTPMeta.CheckResult.UrlfieldMissing"),
                transformMeta);
      } else {
        cr =
            new CheckResult(
                ICheckResult.TYPE_RESULT_ERROR,
                BaseMessages.getString(PKG, "HTTPMeta.CheckResult.UrlfieldOk"),
                transformMeta);
      }

    } else {
      if (Utils.isEmpty(url)) {
        cr =
            new CheckResult(
                ICheckResult.TYPE_RESULT_ERROR,
                BaseMessages.getString(PKG, "HTTPMeta.CheckResult.UrlMissing"),
                transformMeta);
      } else {
        cr =
            new CheckResult(
                ICheckResult.TYPE_RESULT_OK,
                BaseMessages.getString(PKG, "HTTPMeta.CheckResult.UrlOk"),
                transformMeta);
      }
    }
    remarks.add(cr);
  }

  @Override
  public boolean supportsErrorHandling() {
    return true;
  }
}
